#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <asm/uaccess.h>

MODULE_LICENSE("Dual BSD/GPL");

#define DRIVER_NAME "leddev"

#define PIO_0_BASE (*(volatile unsigned char *) 0x80000000)


static int leddev_devs  = 2;
static int leddev_major = 0;
module_param(leddev_major, uint, 0);
static struct cdev leddev_cdev;

struct leddev_data {
    unsigned char val;
    rwlock_t lock;
};  


ssize_t leddev_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
    struct leddev_data *p = filp->private_data;
    unsigned char val;
    int retval = 0;

    printk("%s: count %d pos %lld\n", __func__, count, *f_pos);

    if (count >= 1) {
        if (copy_from_user(&val, &buf[0], 1)) {
            retval = -EFAULT;
            goto out;
        }

        write_lock(&p->lock);
        PIO_0_BASE = val;
        write_unlock(&p->lock);
        retval = count;
    }

out:
    return (retval);
}


static int leddev_open(struct inode *inode, struct file *file)
{
    struct leddev_data *p;

    printk("%s: major %d minor %d (pid %d)\n", __func__,
            imajor(inode),
            iminor(inode),
            current->pid
          );

    p = kmalloc(sizeof(struct leddev_data), GFP_KERNEL);
    if (p == NULL) {
        printk("%s: Not memory\n", __func__);
        return -ENOMEM;
    }

    p->val = 0xff;
    rwlock_init(&p->lock);

    file->private_data = p;

    return 0;
}


static int leddev_close(struct inode *inode, struct file *file)
{
    printk("%s: major %d minor %d (pid %d)\n", __func__,
            imajor(inode),
            iminor(inode),
            current->pid
          );

    if (file->private_data) {
        kfree(file->private_data);
        file->private_data = NULL;
    }

    return 0;
}


static struct file_operations leddev_fops = {
    .write   = leddev_write,
    .open    = leddev_open,
    .release = leddev_close,
};


static int init_leddev(void)
{
    dev_t dev = MKDEV(leddev_major, 0);
    int alloc_ret = 0;
    int major;
    int cdev_err = 0;

    alloc_ret = alloc_chrdev_region(&dev, 0, leddev_devs, DRIVER_NAME);
    if (alloc_ret)
        goto error;
    leddev_major = major = MAJOR(dev);

    cdev_init(&leddev_cdev, &leddev_fops);
    leddev_cdev.owner = THIS_MODULE;

    cdev_err = cdev_add(&leddev_cdev, MKDEV(leddev_major, 0), leddev_devs);
    if (cdev_err) 
        goto error;

    printk(KERN_ALERT "%s driver(major %d) installed.\n", DRIVER_NAME, major);

    return 0;

error:
    if (cdev_err == 0)
        cdev_del(&leddev_cdev);

    if (alloc_ret == 0)
        unregister_chrdev_region(dev, leddev_devs);

    return -1;
}

static void exit_leddev(void)
{
    dev_t dev = MKDEV(leddev_major, 0);

    cdev_del(&leddev_cdev);
    unregister_chrdev_region(dev, leddev_devs);

    printk(KERN_ALERT "%s driver removed.\n", DRIVER_NAME);
}

module_init(init_leddev);
module_exit(exit_leddev);
